// CSE 142, Autumn 2006
// Homework 8 (Critter Safari)
// Author: Marty Stepp
//
// A bunch of methods used to dynamically load critter classes sent across
// the web.  Useful for running 1-on-1 critter tournaments.
//
import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;


public final class ClassUtils {
    // class constants
    public static final String CLASS_EXTENSION = ".class";
    public static final String JAVA_EXTENSION = ".java";
    private static final FileFilter CLASS_FILTER = new ExtensionFilter(CLASS_EXTENSION);

    // Adds 25 of each critter class type to the given model.
    // The only critter-specific code in here; a bit of a cheat
    // so that CritterMain.java doesn't have to have this icky code in it.
    public static void addAllCritterClasses(CritterModel model) {
        for (Class<? extends Critter> critterClass : ClassUtils.getClasses(
                Critter.class, ".")) {
            model.add(25, critterClass);
        }
    }
    
    // An overall method that takes the given file text and dumps it
    // to the disk, renames it, recompiles it, loads the class into the JVM,
    // and returns the associated Class object.
    public static Class<?> doEverything(String fileText, 
            String oldClassName, String newClassName, boolean useTempFolder)
            throws ClassNotFoundException {
        newClassName = sanitizeClassName(newClassName);
        String newFileName = renameAndWriteJavaFile(fileText, 
                oldClassName, newClassName, useTempFolder);
        String classFileName = compile(newFileName);
        new File(newFileName).delete();
        if (useTempFolder) {
            new File(classFileName).deleteOnExit();
        }
        return loadClass(classFileName);
    }
    
    // Compiles the .java source file with the given file name,
    // and returns the file name of the newly compiled .class file.
    // Throws a RuntimeException if the compilation fails.
    public static String compile(String fileName) {
        String folderName = getFolder(fileName);
        String sep = System.getProperty("path.separator");
        String[] args = {"-classpath", "." + sep + folderName, fileName};
        int result = com.sun.tools.javac.Main.compile(args);
        if (result != 0) {
            throw new RuntimeException("Compilation failed: error code " + result);
        }
        
        return removeExtension(fileName) + CLASS_EXTENSION; 
    }
    
    // Downloads the file at the given URL and saves it to the local disk.
    // For example "www.foo.com/bar/baz/Mumble.txt" saves to "Mumble.txt".
    // Returns the file name of the saved local file.
    public static String downloadFile(URL url) throws IOException {
        String fileName = removeFolder(url.getFile());
        if (!(new File(fileName).exists())) {
            OutputStream output = new PrintStream(fileName);
            InputStream input = new BufferedInputStream(url.openStream());
            byte[] buffer = new byte[512000];
            int numRead;
            long numWritten = 0;
            while ((numRead = input.read(buffer)) != -1) {
                output.write(buffer, 0, numRead);
                numWritten += numRead;
            }
            output.close();
        }
        return fileName;
    }
    
    // Loads all classes that extend the given class from the given folder.
    @SuppressWarnings("unchecked")
    public static <T> List<Class<? extends T>> getClasses(Class<T> superClass, 
            String folderName) {
        List<Class<? extends T>> list = new ArrayList<Class<? extends T>>();
        File folder = new File(folderName);
        if (!folder.exists() || !folder.canRead()) {
            // throw new IOException("Can't read folder: " + folderName);
            return list;
        }
        
        for (File file : folder.listFiles(CLASS_FILTER)) {
            String fileName = file.getName();
            if (file.canRead() && !file.isDirectory() && 
                    fileName.endsWith(CLASS_EXTENSION)) {
                try {
                    Class<?> existingClass = Class.forName(removeExtension(fileName));
                    if (!existingClass.isInterface() &&
                            !Modifier.isAbstract(existingClass.getModifiers()) &&
                            superClass.isAssignableFrom(existingClass)) {
                        // then this is a concrete class that implements the interface
                        list.add((Class<? extends T>) existingClass);
                    }
                } catch (Exception e) {
                    System.out.println("error reading " + fileName + ":");
                    e.printStackTrace();
                }
            }
        }
        
        return list;
    }
    
    // Downloads the contents of the .zip file at the given URL, and 
    // returns them as a map from filenames to the bytes of each file.
    public static Map<String, byte[]> getZipFileContents(URL url) throws IOException {
        String fileName = downloadFile(url);
        ZipFile zip = new ZipFile(fileName);
        Map<String, byte[]> zipFilesMap = new TreeMap<String, byte[]>();
        
        // read each file entry from the zip archive
        for (Enumeration<? extends ZipEntry> enu = zip.entries(); enu.hasMoreElements(); ) {
            ZipEntry ze = enu.nextElement();
            if (ze.isDirectory()) {
                continue;
            }
            
            int size = (int) ze.getSize();
            if (size < 0) {
                continue;   // -1 means unknown size.
            }
            
            InputStream input = zip.getInputStream(ze);
            byte[] b = new byte[size];
            int offset = 0;
            while (size - offset > 0) {
                int bytesRead = input.read(b, offset, size - offset);
                if (bytesRead < 0) {
                    break;
                }
                offset += bytesRead;
            }
            
            zipFilesMap.put(ze.getName(), b);
        }
        return zipFilesMap;
    }
    
    // Returns whether the given class implements the given interface.
    public static boolean classImplements(Class<?> clazz, Class<?> interfaceType) {
        for (Class<?> c : clazz.getInterfaces()) {
            if (c == interfaceType) {
                return true;
            }
        }
        return false;
    }
    
    // Dynamically loads the compiled .class file with the given file name
    // into our JVM and returns its Class object.
    // Throws various reflectiony exceptions if the file is bad.
    public static Class<?> loadClass(String fileName) throws ClassNotFoundException {
        String folderName = getFolder(fileName);
        File folder = new File(folderName);
        ClassLoader loader = ClassLoader.getSystemClassLoader();
        
        ClassLoader urlLoader = loader;
        try {
            URL fileUrl = new URL("file:" + System.getProperty("user.dir") + 
                    File.separator + fileName);
            
            File currentDir = new File(System.getProperty("user.dir"));
            urlLoader = URLClassLoader.newInstance(new URL[] {
                    folder.toURL(), currentDir.toURL(), fileUrl}, loader);
        } catch (MalformedURLException mfurle) {
            mfurle.printStackTrace();   // this will never happen
        }
        
        String className = removeExtension(removeFolder(fileName));
        Class<?> clazz = urlLoader.loadClass(className);
        return clazz;
    }
    
    // Reads the given file's text fully and returns it as a String.
    public static String readEntireFile(String fileName) throws IOException {
        BufferedReader reader = new BufferedReader(new FileReader(fileName));
        StringBuilder text = new StringBuilder();
        while (reader.ready()) {
            text.append((char) reader.read());
        }
        return text.toString();
    }
    
    // Treats fileText as the text of a Java source file, and
    // replaces occurrences of its class name with the given new class name,
    // then writes it to a new .java file with that name.
    public static String renameAndWriteJavaFile(String fileText, 
            String oldClassName, String newClassName, boolean useTempFolder) {
        // replace the class name in the code
        fileText = fileText.replaceAll(oldClassName, newClassName);
        
        // write the modified text to a new file
        String newFileName = newClassName + JAVA_EXTENSION;
        if (useTempFolder) {
            newFileName = System.getProperty("java.io.tmpdir") + newFileName;
        }
        
        try {
            PrintStream output = new PrintStream(newFileName);
            output.print(fileText);
            output.close();
        } catch (FileNotFoundException fnfe) {
            fnfe.printStackTrace();  // this will never happen; we're writing the file!
        }
        
        return newFileName;
    }
    
    // Removes any characters from given text that wouldn't be acceptable 
    // in a Java class name.
    // Not perfect (e.g., doesn't prevent names that start with a number).
    public static String sanitizeClassName(String text) {
        text = text.replaceAll("[^A-Za-z0-9_$]+", "_");
        return text;
    }
    
    public static void writeBytes(byte[] bytes, String fileName) throws IOException {
        FileOutputStream output = new FileOutputStream(fileName);
        output.write(bytes);
        output.close();
    }
    
    // pre: no folders in fileName (no "foo/bar/Baz.java")
    private static String getFolder(String fileName) {
        int slash = fileName.lastIndexOf(File.separatorChar);
        if (slash < 0) {
            slash = fileName.lastIndexOf("/");  // fallback
        }
        
        if (slash >= 0) {
            return fileName.substring(0, slash + 1);
        } else {
            return "./";
        }
    }
    
    // pre: no folders in fileName (no "foo/bar/Baz.java")
    private static String removeExtension(String fileName) {
        int dot = fileName.lastIndexOf(".");
        if (dot >= 0) {
            fileName = fileName.substring(0, dot);
        }
        return fileName;
    }
    
    // pre: no folders in fileName (no "foo/bar/Baz.java")
    private static String removeFolder(String fileName) {
        int slash = fileName.lastIndexOf(File.separatorChar);
        if (slash < 0) {
            slash = fileName.lastIndexOf("/");  // fallback
        }

        if (slash >= 0) {
            fileName = fileName.substring(slash + 1);
        }
        return fileName;
    }    

    // inner class to filter files by extension
    public static class ExtensionFilter implements FileFilter {
        private String extension;
        
        public ExtensionFilter(String extension) {
            this.extension = extension;
        }
        
        public boolean accept(File f) {
            return f != null && f.exists() && f.canRead() &&
                    f.getName().endsWith(extension);
        }
    }
}
