// CSE 142, Autumn 2006
// Homework 8 (Critter Safari)
// Authors: Marty Stepp, Stuart Reges
//
// The overall graphical user interface for the Critter simulation.
//
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;

public final class CritterGui implements ActionListener, ChangeListener, Observer, WindowListener {
    // class constants
    private static final long serialVersionUID = 0;
    private static final String TITLE = "CSE 142 Critter Tournament";
    private static final int DELAY = 100;   // default MS between redraws
    private static final int DEFAULT_NUMBER_OF_CRITTERS = 25;  // # of critters to add
    private static final int MAX_CLASS_NAME_LENGTH = 12;

    // constants for saving/loading GUI state
    private static final boolean SHOULD_SAVE_SETTINGS = true;
    private static final String LAST_HOST_NAME_KEY = "lastHostName";
    private static final String FPS_KEY = "fps";
    private static final String ALWAYS_ACCEPT_KEY = "alwaysAccept";
    private static final String SAVE_STATE_FILE_NAME = "_hw8_gui_saved_settings.txt";
    
    // constant for loading files from the course web site
    private static final String CODE_BASE = "http://www.cs.washington.edu/education/courses/cse142/06au/homework/student_critters.zip";
    
    // Helper method to create a JButton with the given properties.
    public static JButton createButton(String text, char mnemonic,
            ActionListener listen, Container panel) {
        JButton button = new JButton(text);
        if (mnemonic != '\0') {
            button.setMnemonic(mnemonic);
        }
        button.addActionListener(listen);
        if (panel != null) {
            panel.add(button);
        }
        return button;
    }
    
    // Helper method to create a JCheckBox with the given properties.
    public static JCheckBox createCheckBox(String text, char mnemonic,
            ActionListener listen, Container panel) {
        JCheckBox box = new JCheckBox(text);
        if (mnemonic != '\0') {
            box.setMnemonic(mnemonic);
        }
        box.addActionListener(listen);
        if (panel != null) {
            panel.add(box);
        }
        return box;
    }
    
    // Helper method to create a JSlider with the given properties.
    public static JSlider createSlider(int min, int max, int initial,
            int majorTick, int minorTick, ChangeListener listen, Container panel) {
        JSlider slider = new JSlider(min, max, initial);
        slider.setMajorTickSpacing(majorTick);
        slider.setMinorTickSpacing(minorTick);
        slider.setSnapToTicks(true);
        slider.setPaintTicks(true);
        // slider.setPaintLabels(true);
        Dimension size = slider.getPreferredSize();
        slider.setPreferredSize(new Dimension(size.width / 2, size.height));
        slider.addChangeListener(listen);
        if (panel != null) {
            panel.add(slider);
        }
        return slider;
    }
    
    // fields
    private CritterModel model;
    private CritterPanel panel;
    private JFrame frame;
    private javax.swing.Timer timer;
    private JButton go, stop, tick, reset, loadFromWeb;
    private JSlider slider;
    private JComponent east;
    private JCheckBox alwaysAccept;
    private String lastHostName = "";
    private JLabel moves;

    // receives critters from others and lets me voluntarily send mine out
    // packets = [hostname, classname, critter text]
    private CritterNetworkManager networkSenderListener;
    
    // lets me host my wolf so others can reach out and request it from me
    // packets = [hostname, classname requested]
    private CritterNetworkManager networkServer;
    
    // keeps track of which classes already have a ClassPanel on the east side
    // of the window, so we know when we need to add a new one (on network receive etc)
    private Map<String, ClassPanel> counts;

    // Constructs a new GUI to display the given model of critters.
    public CritterGui(CritterModel model) {
        this.model = model;
        model.addObserver(this);
        
        // try to load settings from disk (fail silently)
        Properties props = null;
        if (SHOULD_SAVE_SETTINGS) {
            try {
                props = loadConfiguration();
            } catch (IOException ioe) {}
        }
        
        // set up network listeners
        networkSenderListener = new CritterNetworkManager();
        networkSenderListener.getReceiveEvent().addObserver(this);
        networkSenderListener.getErrorEvent().addObserver(this);
        networkServer = new CritterNetworkManager(networkSenderListener.getPort() + 1);
        networkServer.getReceiveEvent().addObserver(this);
        networkServer.getErrorEvent().addObserver(this);
        
        // set up critter picture panel and set size
        panel = new CritterPanel(model);
        panel.setBorder(BorderFactory.createLineBorder(Color.BLACK));
        
        // add the animation timer
        timer = new javax.swing.Timer(DELAY, this);
        timer.setCoalesce(true);
        
        // east panel to store critter class info
        counts = new TreeMap<String, ClassPanel>();
        east = new JPanel(new GridLayout(0, 1));
        east.setBorder(BorderFactory.createTitledBorder("Classes:"));
        loadFromWeb = createButton("Load from Web...", 'L', this, east);
        
        // timer controls
        JPanel southcenter = new JPanel();
        go = createButton("Go", 'G', this, southcenter);
        go.setBackground(Color.GREEN);
        stop = createButton("Stop", 'S', this, southcenter);
        stop.setBackground(new Color(255, 96, 96));
        tick = createButton("Tick", 'T', this, southcenter);
        tick.setBackground(Color.YELLOW);
        reset = createButton("Reset", 'R', this, southcenter);
        
        // slider for animation speed
        JPanel southwest = new JPanel();
        southwest.add(new JLabel("Speed:"));
        slider = createSlider(1, 41, 1000 / DELAY, 20, 5, this, southwest);
        moves = new JLabel("0 moves");
        southwest.add(moves);
        
        // checkbox 
        JPanel southeast = new JPanel(new FlowLayout(FlowLayout.CENTER, 20, 5));
        alwaysAccept = createCheckBox("Always Accept", 'A', this, southeast);
        alwaysAccept.setToolTipText("When checked, automatically accepts critters sent to you " +
                "and automatically shares requested critters.");
        
        // south panel to hold various widgets
        Container south = new JPanel(new BorderLayout());
        south.add(southcenter);
        south.add(southwest, BorderLayout.WEST);
        south.add(southeast, BorderLayout.EAST);
        
        JPanel center = new JPanel();
        center.add(panel);

        // use saved settings, if any (fail silently)
        if (props != null) {
            try {
                slider.setValue(Integer.parseInt(props.getProperty(FPS_KEY)));
                alwaysAccept.setSelected(Boolean.parseBoolean(props.getProperty(ALWAYS_ACCEPT_KEY)));
                lastHostName = props.getProperty(LAST_HOST_NAME_KEY, "");
                // timer.setDelay(Integer.parseInt(props.getProperty(FPS_KEY)));
            } catch (Exception e) {}
        }
        
        // create frame and do layout
        frame = new JFrame();
        frame.setTitle(TITLE + ": " + CritterNetworkManager.getHostName() + " " + 
                CritterNetworkManager.getIpAddresses());
        // frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.addWindowListener(this);
        // frame.setResizable(false);
        frame.add(center, BorderLayout.CENTER);
        frame.add(south, BorderLayout.SOUTH);
        frame.add(east, BorderLayout.EAST);
    }
    
    // Responds to action events in the GUI.
    public void actionPerformed(ActionEvent e) {
        Object src = e.getSource();
        if (src == go) {
            timer.start();
        } else if (src == stop) {
            timer.stop();
        } else if (src == timer || (src == tick && !timer.isRunning())) {
            model.update();
        } else if (src == reset) { 
            model.reset();
        } else if (src == loadFromWeb) {
            new Thread(new ZipDownloader()).start();
        }
    }
    
    // Responds to change events on the slider.
    public void stateChanged(ChangeEvent e) {
        int fps = slider.getValue();
        timer.setDelay(1000 / fps);
    }
    
    // Starts the simulation. Assumes all critters have already been added.
    public void start() {
        setCounts();
        
        frame.pack();
        frame.setVisible(true);
        
        // start network listeners
        try {
            networkSenderListener.start();
            networkServer.start();
        } catch (java.net.BindException e) {
            JOptionPane.showMessageDialog(frame,
                    "Error: The network is already in use.\n\n" + 
                    "If you want to be able to send and receive critters over the network,\n" +
                    "please close all instances of the Critter Safari and run it again.",
                    "Network in use", JOptionPane.ERROR_MESSAGE);
        } catch (IOException e) {
            JOptionPane.showMessageDialog(frame,
                    "Error starting network listener:\n" + e, "Network error",
                    JOptionPane.ERROR_MESSAGE);
            e.printStackTrace();
        }
    }
    
    // Responds to Observable updates in the model.
    public void update(Observable o, Object arg) {
        if (o == model) {
            // model is notifying us of an update
            if (arg == CritterModel.Event.ADD || arg == CritterModel.Event.REMOVE_ALL) {
                setCounts();
                moves.setText(model.getMoveCount() + " moves");
            }
            // TODO: remove overall gui as observer of model?
        } else if (o == networkSenderListener.getReceiveEvent() && arg != null) {
            // we received a message (a class to load)
            String[] strings = (String[]) arg;
            load(strings);
            // setCounts();
        } else if (o == networkServer.getReceiveEvent() && arg != null) {
            // we received a message (a request to send our wolf)
            String[] strings = (String[]) arg;
            String hostName = strings[0];
            String className = strings[1];

            if (!alwaysAccept.isSelected()) {
                // "Always Accept" not checked, so ask to confirm
                int choice = JOptionPane.showConfirmDialog(frame, "Host \"" + hostName +
                        "\" requests your " + className + " class.  Send it?",
                        "Critter send request",
                        JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
                if (choice != JOptionPane.YES_OPTION) {
                    // refuse request (send back a null answer)
                    sendClass(null, hostName);
                    return;
                }
            }
            
            // user confirmed, so send the class data
            sendClass(className, hostName);
        } else if (o == networkSenderListener.getErrorEvent() ||
                   o == networkServer.getErrorEvent()) {
            // something failed in a network send/receive attempt
            Exception e = (Exception) arg;
            String message;
            if (e instanceof UnknownHostException) {
                message = "Cannot reach target computer:\n\n" + e;
            } else if (e instanceof ConnectException) {
                message = "Target computer refused connection:\n\n" + e;
            } else if (e instanceof IOException) {
                message = "I/O error:\n" + e;
            } else {
                message = e.toString();
            }
            JOptionPane.showMessageDialog(frame, message, "network error",
                    JOptionPane.ERROR_MESSAGE);
            e.printStackTrace();
        }
    }
    
    // Called when the window is about to close.
    // Used to save the GUI's settings.
    public void windowClosing(WindowEvent e) {
        if (SHOULD_SAVE_SETTINGS) {
            try {
                saveConfiguration();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
        System.exit(0);
    }
    
    // Required to implement WindowListener interface.
    public void windowActivated(WindowEvent e) {}
    public void windowClosed(WindowEvent e) {}
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
    public void windowOpened(WindowEvent e) {}
    
    
    // Helper method to read an integer input from a set of choices.
    private int getInput(String message, Object defaultValue, Object... choices) {
        Object countStr = JOptionPane.showInputDialog(frame, message, "Question",
                JOptionPane.QUESTION_MESSAGE, null, choices, defaultValue);
        if (countStr == null) {
            return -1;
        }
        try {
            return Integer.parseInt(countStr.toString());
        } catch (NumberFormatException e) {
            return -1;
        }
    }
    
    // Helper method to read a String input (with the given initial String in the field)
    // and return a default value if an empty string is entered.
    private String getInput(String message, String initialValue, String defaultValue) {
        String input = JOptionPane.showInputDialog(frame, message, "Question", JOptionPane.QUESTION_MESSAGE, null, null, initialValue).toString();
        if (input != null && input.length() == 0) {
            input = defaultValue;
        }
        return input;
    }
    
    // Returns whether the given class is one that came from the network.
    private boolean isNetworkClass(String className) {
        return className.indexOf('_') >= 0;
    }
    
    // Loads a class received over the network.
    // The given array contains [host name, class name, class text]
    @SuppressWarnings("unchecked")
    private void load(String[] strings) {
        String hostName = strings[0];
        String oldClassName = strings[1];
        String fileText = strings[2];
        
        if (fileText == null) {
            // they refused our request and sent back a null
            JOptionPane.showMessageDialog(frame, hostName + " refused the request.", 
                    "Request refused", JOptionPane.ERROR_MESSAGE);
            return;
        }
        
        // rename class, such as "Wolf" -> "Wolf_stepp"
        String newClassName = oldClassName + "_" + hostName;
        
        // find out how many critters to add to the world
        int count = DEFAULT_NUMBER_OF_CRITTERS;
        if (!alwaysAccept.isSelected()) {
            count = getInput("Received " + oldClassName + " from host \"" + hostName + 
                    "\".\nHow many to add? (Or Cancel to refuse this class)", 
                    DEFAULT_NUMBER_OF_CRITTERS,
                    0, 1, 25, 50, 100);
            if (count < 0) {
                return;
            }
        }
        
        // try to compile and load the received class, add it to simulation
        try {
            Class<? extends Critter> critterClass = (Class<? extends Critter>) ClassUtils
                    .doEverything(fileText, oldClassName, newClassName, false);
            model.add(count, critterClass);
        } catch (CritterModel.TooManyCrittersException e) {
            JOptionPane.showMessageDialog(frame,
                    "Error: Not enough room to add all critters",
                    "Too many critters", JOptionPane.ERROR_MESSAGE);
        } catch (CritterModel.InvalidCritterClassException e) {
            JOptionPane.showMessageDialog(frame,
                    "Problem with critter class:\n" + e + "\n\n" +
                    "This is probably DrJava's fault; DrJava has some issues with dynamically loading code over the network.\n" +
                    "If you've got a " + newClassName + ".class file in your program's folder, the file got to you successfully, but\n" +
                    "DrJava wasn't able to load it into the simulator.\n\n" +
                    "Try closing and re-running the simulator; this will give DrJava a chance to see the new animal type and load it.\n" +
                    "Hopefully, when you re-run the simulator, the " + newClassName + " type will appear.",
                    
                    "Problem with critter class",
                    JOptionPane.ERROR_MESSAGE);
        } catch (ClassNotFoundException cnfe) {
            JOptionPane.showMessageDialog(frame, "Error loading class:\n" + cnfe, "Error",
                    JOptionPane.ERROR_MESSAGE);
            cnfe.printStackTrace();
        }
    }
    
    private Properties loadConfiguration() throws IOException {
        Properties prop = new Properties();
        prop.load(new FileInputStream(SAVE_STATE_FILE_NAME));
        return prop;
    }
    
    private void saveConfiguration() throws IOException {
        Properties prop = new Properties();
        int fps = slider.getValue();
        prop.setProperty(LAST_HOST_NAME_KEY, lastHostName);
        prop.setProperty(FPS_KEY, String.valueOf(fps));
        prop.put(ALWAYS_ACCEPT_KEY, String.valueOf(alwaysAccept.isSelected()));
        prop.store(new PrintStream(SAVE_STATE_FILE_NAME), "this is a comment");
    }
    
    // sends the given class code to the given host computer
    // if className is null, sends null text to signify request refused
    private void sendClass(String className, String hostName) {
        try {
            String fileText = null;
            if (className != null) {
                fileText = ClassUtils.readEntireFile(className + ".java");
            }
            networkSenderListener.sendClass(hostName, className, fileText);
        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(frame,
                    "Error reading file:\n" + ioe, "I/O Error",
                    JOptionPane.ERROR_MESSAGE);
        }
    }
    
    // Adds right-hand column of labels showing how many of each type are alive.
    // Updates the counter labels to store the current count information.
    private void setCounts() {
        Set<String> classNames = model.getClassNames();
        if (classNames.size() == counts.size()) {
            return;   // nothing to do
        }
        
        for (ClassPanel cpanel : counts.values()) {
            east.remove(cpanel);
        }
        counts.clear();
        for (String className : classNames) {
            ClassPanel cpanel = new ClassPanel(className);
            east.add(cpanel);
            counts.put(className, cpanel);
        }
        east.validate();
        frame.pack();
    }
    
    // One of the east panels representing a critter class.
    private class ClassPanel extends JPanel implements ActionListener, Observer {
        private static final long serialVersionUID = 0;
        
        // fields
        private String className;
        private JButton send, request, delete;

        // Constructs a new ClassPanel to hold info about the given critter class.
        public ClassPanel(String className) {
            this.className = className;
            updateBorder();
            model.addObserver(this);

            if (isNetworkClass(className)) {
                setLayout(new BorderLayout(0, 0));
                delete = createButton("Delete", '\0', this, null);
                delete.setForeground(Color.RED.darker());
                delete.setToolTipText("Delete class " + className);
                add(delete, BorderLayout.NORTH);
            } else {
                setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
                send = createButton("Send", '\0', this, this);
                request = createButton("Request", '\0', this, this);
            }
        }
        
        // Handles action events in this panel.
        public void actionPerformed(ActionEvent e) {
            Object src = e.getSource();
            if (src == send) {
                if (isNetworkClass(className)) {
                    return;  // skip network classes
                }
                
                // ask what computer to send to
                String hostName = getInput("Send your " + className + 
                        " to what computer name or IP address?\n" +
                        "(or leave blank to connect to local host)\n ", lastHostName, "localhost");
                if (hostName != null) {
                    // send class to computer
                    lastHostName = (hostName.equals("localhost") ? "" : hostName);
                    sendClass(className, hostName);
                }
            } else if (src == request) {
                // ask what computer to request from
                String hostName = getInput("Request " + className
                        + " from what computer name or IP address?\n"
                        + "(or leave blank to request from local host)\n ", lastHostName, "localhost");
                if (hostName != null) {
                    // request class from computer
                    lastHostName = (hostName.equals("localhost") ? "" : hostName);
                    networkServer.requestClass(hostName, className);
                }
            } else if (src == delete) {
                // delete this network class from the system
                String classFileName = className + ClassUtils.CLASS_EXTENSION;
                int choice = JOptionPane.showConfirmDialog(frame, "Delete all " + 
                        className + " critters and remove the " + classFileName +
                        " file from your disk?",
                        "Delete class?", JOptionPane.YES_NO_OPTION, 
                        JOptionPane.QUESTION_MESSAGE);
                if (choice != JOptionPane.YES_OPTION) {
                    return;
                }
                
                model.removeAll(className);
                
                // set the .class file to be deleted when the VM exits
                // BUG: if they then send us that same file while GUI is running,
                // I'm unable to cancel the delete-on-exit request and it'll
                // get deleted when you close your GUI.  Oh well.
                new File(classFileName).deleteOnExit();
            }
        }
        
        // Handles Observable updates from the model.
        public void update(Observable o, Object arg) {
            updateBorder();
        }
        
        // Updates the border text to 
        private void updateBorder() {
            setBorder(BorderFactory.createTitledBorder(trimClassName() + "=" + model.getCount(className)));
        }
        
        private String trimClassName() {
            if (className.length() <= MAX_CLASS_NAME_LENGTH) {
                return className;
            } else {
                return className.substring(0, MAX_CLASS_NAME_LENGTH) + "~";
            }
        }
    }
    
    private class ListOptionPane extends JDialog implements ActionListener {
        private static final long serialVersionUID = 0;
        private JList list;
        private JButton ok, cancel;
        
        public ListOptionPane(Collection<String> items) {
            super(frame, true);
            setTitle("Load...");
            
            list = new JList(items.toArray());
            list.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
            list.setVisibleRowCount(Math.min(12, items.size()));
            
            JPanel south = new JPanel();
            ok = createButton("OK", 'O', this, south);
            cancel = createButton("Cancel", 'C', this, south);

            add(new JScrollPane(list));
            add(south, BorderLayout.SOUTH);
            pack();
            setLocation(frame.getX() + (frame.getWidth() - getWidth()) / 2,
                    frame.getY() + (frame.getHeight() - getHeight()) / 2);
        }
        
        public void actionPerformed(ActionEvent e) {
            Object src = e.getSource();
            if (src == ok) {
                
            } else if (src == cancel) {
                setVisible(false);
                // dispose();
            }
            setVisible(false);
        }
        
        public Object[] getSelectedValues() {
            return list.getSelectedValues();
        }
    }
    
    private class ZipDownloader implements Runnable {
        @SuppressWarnings("unchecked")
        public void run() {
            loadFromWeb.setEnabled(false);
            try {
                Map<String, byte[]> zipFilesMap = ClassUtils.getZipFileContents(new URL(CODE_BASE));
                
                ListOptionPane dialog = new ListOptionPane(zipFilesMap.keySet());
                dialog.setVisible(true);
                Object[] selectedItems = dialog.getSelectedValues();
                System.out.println(Arrays.toString(selectedItems));
                
                // write the selected classes as files to disk
                for (Object item : selectedItems) {
                    String fileName = item.toString();
                    byte[] bytes = zipFilesMap.get(fileName);
                    ClassUtils.writeBytes(bytes, fileName);
                    Class<? extends Critter> critterClass = 
                        (Class<? extends Critter>) ClassUtils.loadClass(fileName);
                    model.add(DEFAULT_NUMBER_OF_CRITTERS, critterClass);
                }
            } catch (IOException e) {
                JOptionPane.showMessageDialog(frame,
                        "Error downloading ZIP data:\n" + e, "I/O error",
                        JOptionPane.ERROR_MESSAGE);
                e.printStackTrace();
            } catch (Exception e) {
                JOptionPane.showMessageDialog(frame,
                        "Error loading class data:\n" + e, "Class loading error",
                        JOptionPane.ERROR_MESSAGE);
                e.printStackTrace();
            }
            loadFromWeb.setEnabled(true);
        }
    }
}
