// CSE 142, Autumn 2006
// Homework 8 (Critter Safari)
// Author: Marty Stepp
//
// Handles network responsibilities in the critter game.
//
import java.io.*;
import java.net.*;
import java.util.*;

// Critter-specific usage of the network manager object.
public class CritterNetworkManager extends NetworkManager {
    // class constants
    private static final int DEFAULT_PORT = 5142;

    // Constructs a CritterNetworkManager at the default port.
    public CritterNetworkManager() {
        super(DEFAULT_PORT);
    }
    
    // Constructs a CritterNetworkManager at the given port.
    public CritterNetworkManager(int port) {
        super(port);
    }
    
    // Sends out a message to the given host requesting that they send
    // their class with the given name back to us.
    public void requestClass(String host, String className) {
        send(host, getHostName(), className);
    }
    
    // Sends out the text of the given class to the given host.
    public void sendClass(String host, String className, String fileText) {
        send(host, getHostName(), className, fileText);
    }
}

// General reusable version of event-driven network manager object.
class NetworkManager {
    // class constants
    private static final boolean DEBUG = false;
    private static String hostName;
    private static String ipAddress;
    private static String ipAddresses = "";
    
    static {
        // use host name to name classes sent over the wire
        // also grab IP address(es) of local computer
        try {
            InetAddress localhost = InetAddress.getLocalHost();
            hostName = localhost.getHostName();
            ipAddress = localhost.getHostAddress();
            
            for (InetAddress addr : InetAddress.getAllByName(hostName)) {
                ipAddresses += " " + addr.getHostAddress();
            }
            ipAddresses = ipAddresses.trim();
        } catch (UnknownHostException e) {
            hostName = "unknown_host";
            ipAddress = ipAddresses = "127.0.0.1";
        }
    }

    // Returns this computer's host name.
    public static String getHostName() {
        return hostName;
    }
    
    // Returns this computer's IP address.
    public static String getIpAddress() {
        return ipAddress;
    }
    
    // Returns this computer's IP address.
    public static String getIpAddresses() {
        return ipAddresses;
    }

    // fields
    private int port;
    private boolean shouldContinue;
    private List<Message> outQueue;   // queue of messages to send
    private Thread sendThread, receiveThread;
    private Event<String[]> receive;  // events to represent messages received
    private Event<Exception> error;   // and errors that occur
    
    // Constructs a network manager at the given port.
    public NetworkManager(int port) {
        this.port = port;
        shouldContinue = true;
        outQueue = Collections.synchronizedList(new LinkedList<Message>());
        receive = new Event<String[]>();
        error = new Event<Exception>();
    }
    
    // Returns this network manager's observable error event object,
    // so observers can be notified about errors and respond to them.
    public Event<? extends Exception> getErrorEvent() {
        return error;
    }
    
    // Returns this network manager's port.
    public int getPort() {
        return port;
    }
    
    // Returns this network manager's observable receive event object,
    // so observers can be notified when messages are received.
    public Event<String[]> getReceiveEvent() {
        return receive;
    }
    
    // Sends a message to the given host containing the given strings.
    public void send(String host, String... strings) {
        synchronized (outQueue) {
            Message message = new Message(host, strings);
            outQueue.add(message);
        }
    }
    
    // Notifies this network manager that you want it to shut down
    // and stop listening for messages.
    public void stop() {
        shouldContinue = false;
    }
    
    // Starts listening for messages on the network.
    public void start() throws IOException {
        shouldContinue = true;
        if (sendThread == null) {
            sendThread = new Thread(new SendRunnable());
            receiveThread = new Thread(new ReceiveRunnable());
        }

        sendThread.start();
        receiveThread.start();
    }

    // A helper class to represent observable events.
    public class Event<T> extends Observable {
        // Convenience method to replace notifyObservers.
        public void fire() {
            notifyObservers(null);
        }
        
        // Convenience method to replace notifyObservers.
        public void fire(T arg) {
            setChanged();
            super.notifyObservers(arg);
        }
    }
    
    // Listening thread that waits for messages to arrive.
    private class ReceiveRunnable implements Runnable {
        // fields
        private ServerSocket srvSock;

        public ReceiveRunnable() throws IOException {
            srvSock = new ServerSocket(port);
        }

        // Runs the listening thread.
        public void run() {
            try {
                while (shouldContinue) {
                    // wait for a message to arrive
                    Socket sock = srvSock.accept();
                    InputStream stream = sock.getInputStream();
                    ObjectInputStream ois = new ObjectInputStream(stream);
                    
                    // read message
                    String[] strings = (String[]) ois.readObject();
                    if (DEBUG) System.out.println("Received on port " + port + ":\n" + Arrays.toString(strings));
                    
                    // notify observers that message has arrived
                    receive.fire(strings);
                }
            } catch (Exception e) {
                // notify observers that an error occurred
                error.fire(e);
            }
        }
    }
    
    // Sending thread that outputs messages from the output queue to the network.
    private class SendRunnable implements Runnable {
        public void run() {
            while (shouldContinue) {
                // test-and-test-and-set paradigm
                if (!outQueue.isEmpty()) { 
                    synchronized (outQueue) {
                        if (!outQueue.isEmpty()) {
                            // grab message from the queue
                            Message message = outQueue.remove(0);
                            
                            try {
                                // send the message
                                Socket sock = new Socket(message.host, port);
                                OutputStream stream = sock.getOutputStream();
                                ObjectOutputStream oos = new ObjectOutputStream(stream);
                                oos.writeObject(message.strings);
                                if (DEBUG) System.out.println("Sent on port " + port + ":\n" + message);
                            } catch (IOException e) {
                                // notify observers that an error occurred
                                error.fire(e);
                            }
                        }
                    }
                }
                
                // wait 1 second between checks for messages so this thread
                // won't drain 100% of the CPU
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {}
            }
        }
    }
    
    // Convenience classto represent messages about to be sent over the network. 
    private static class Message implements Serializable {
        // class constants
        private static final long serialVersionUID = 0;
        
        // fields
        public String host;
        public String[] strings;
        
        // Constructs a message to the given host containing the given strings.
        public Message(String host, String[] strings) {
            this.host = host;
            this.strings = strings;
        }
        
        // Returns debug text about this message.
        public String toString() {
            return "Message{host=" + host + ", strings=" + Arrays.toString(strings) + "}";
        }
    }
}
